/**************/
/* Variency.c */
/********/
/* Thomas RIBO, december 2001, DESS IIN. */
/* Code licensed under the terms of the General Public License
   http://www.gnu.org */

/********/
/* Plugin for texture detection. This one computes the variencies for
   each pixel in the image. If the variency is greater than a given
   value, we assume this pixel is part of a texture. */

#include "Variency.h"
#include <stdio.h>
#include <stdlib.h>

/* This is needed by a Gimp plugin */
MAIN()

/* The famous query function */
static void query(void)
{
   static GimpParamDef args[] =
   {
	  {GIMP_PDB_INT32, "run_mode", "Interactive"},
	  {GIMP_PDB_IMAGE, "imageID", "Input image"},
	  {GIMP_PDB_DRAWABLE, "drawableID", "Input drawable"},
   };

   static GimpParamDef return_vals[] = {
	  {GIMP_PDB_IMAGE, "image", "Output image"},
   };

   static int nb_args =  sizeof (args) / sizeof (args[0]);
   static int nb_return_vals = sizeof (return_vals) / sizeof (return_vals[0]);

   gimp_install_procedure
	  ("plug_in_variency_texture_detection",
	   "Find textured areas in an image, by the variency method",
	   "Find textured areas in an image, by the variency method",
	   "Thomas RIBO",
	   "Thomas RIBO",
	   "2001",
	   "<Image>/Texture/Detection/Variency>",
	   "GRAY",
	   GIMP_PLUGIN,
	   nb_args, nb_return_vals, args, return_vals);
}

/* The method that is ran when the plugin is called */
static void run(char* name,
	int nb_params,
	GimpParam* param,
	int* nb_return_vals,
	GimpParam** return_vals)
{
   GimpDrawable* outputDrawable;
   GimpDrawable* inputDrawable;
   gint32 outputImageID;
   gint32 outputLayerID;

   /* Gets the input image */
   inputDrawable = gimp_drawable_get(param[2].data.d_drawable);

   /* Create the output image */
   outputImageID = gimp_image_new(inputDrawable->width,
								  inputDrawable->height,
								  GIMP_GRAY);
   /* Create a layer on the output image */
   outputLayerID = gimp_layer_new(outputImageID,
								  "Background",
								  inputDrawable->width,
								  inputDrawable->height,
								  GIMP_GRAY_IMAGE,
								  100.0,		/* opacity */
								  GIMP_NORMAL_MODE);
   gimp_image_add_layer(outputImageID, outputLayerID, 0);
   /* Create a drawable for the output image */
   outputDrawable = gimp_drawable_get(outputLayerID);

//   fprintf(stderr, "Starting Variency plugin...\n");

   /* Let the function do the work for us */
   variencyMethod(inputDrawable, outputDrawable);

   /* Displays the result */
   gimp_display_new(outputImageID);
}

/* FIXME: give the value in parameter of the function */
#define MAX_VARIENCY 256.0
/* FIXME: give the window size in parameter of the function */
#define WINDOW_SIZE 5

static void variencyMethod(GimpDrawable* inputDrawable,
						   GimpDrawable* outputDrawable)
{
   GimpPixelRgn inputRegion;
   GimpPixelRgn outputRegion;

   guchar blackValue = 255;
   guchar whiteValue = 0;

   /* FIXME: if the image size is lessier than the window size, the subtractions will make us problems */
   unsigned startRow = (unsigned) WINDOW_SIZE /2;
   unsigned endRow = inputDrawable->height - (unsigned) WINDOW_SIZE /2;
   unsigned startCol = (unsigned) WINDOW_SIZE /2;
   unsigned endCol = inputDrawable->width - (unsigned) WINDOW_SIZE /2;
   unsigned row = startRow;
   unsigned col = startCol;

//   fprintf(stderr, "Creating pixel regions...\n");

   /* Region creation */
   gimp_pixel_rgn_init(&inputRegion, inputDrawable,
					   0, 0,
					   inputDrawable->width, inputDrawable->height,
					   FALSE, FALSE);
   gimp_pixel_rgn_init(&outputRegion, outputDrawable,
					   0, 0,
					   outputDrawable->width, outputDrawable->height,
					   TRUE, TRUE);

//   fprintf(stderr, "Ok, ready to parse the image...\n");

   /* Ok, let's go... */
   for(row = startRow; row <= endRow; row++)
   {
	  for(col = startCol; col <= endCol; col++)
	  {
		 /* Creates the window */
		 unsigned startColWindow = col - (unsigned) WINDOW_SIZE /2 -1;
		 unsigned startRowWindow = row - (unsigned) WINDOW_SIZE /2 -1;
		 GimpPixelRgn window;
		 gimp_pixel_rgn_init(&window, inputDrawable,
							 startColWindow, startRowWindow,
							 WINDOW_SIZE, WINDOW_SIZE,
							 FALSE, FALSE);

		 /* Computes the variency and determine whether the pixel is
            part of a region or not. Colorify the output image */
		 if(variency(&window) > MAX_VARIENCY)
		 {
			fprintf(stderr, "Setting 0 in pixel %d,%d\n", col, row);
			gimp_pixel_rgn_set_pixel(&outputRegion, &blackValue, col, row);
		 }
		 else
		 {
			gimp_pixel_rgn_set_pixel(&outputRegion, &whiteValue, col, row);
		 }
	  }
   }
}

/* Computes the average of pixel levels on a region. This computes
   only in gray-levels for the moment */
static double average(GimpPixelRgn* region)
{
   unsigned sum = 0;
   unsigned y, x;

   guchar* data = g_new(guchar, region->h * region->w);

   /* Get the region data */
   gimp_pixel_rgn_get_rect(region, data,
						   region->x, region->y,
						   region->w, region->h);

   /* Sum of pixel levels */
   for(y = 0; y < region->h; y++)
   {
	  for(x = 0; x < region->w; x++)
	  {
		 guchar value = data[y * region->w + x];
		 sum += value;
	  }
   }

   g_free(data);

   /* Computes and return the average */
   return ((double) sum) / ((double) region->h * region->w);
}

/* Computes the variency ("variance" in french) of pixel levels on a
   region. This computes only in gray-levels for the moment */
static double variency(GimpPixelRgn* region)
{
   double variencySum = 0.0;
   unsigned y, x;

   double averageValue = average(region);

   guchar* data = g_new(guchar, region->h * region->w);

   /* Get the region data */
   gimp_pixel_rgn_get_rect(region, data,
						   region->x, region->y,
						   region->w, region->h);

   /* Computes the variency */
   for(y = 0; y < region->h; y++)
   {
	  for(x = 0; x < region->w; x++)
	  {
		 guchar value = data[y * region->w + x];
		 variencySum += (value - averageValue) * (value - averageValue);
	  }
   }

   g_free(data);

   return ((double) variencySum) / ((double) region->h * region->w);
}
